package algod;

import java.util.Vector;
import java.util.Random;
import graphelastic.GraphAnimFrame;

/**
 * Gnre et stocke un ensemble de sommets d'un graphe.
 * Peut afficher les sommets de manire graphique grce  un <CODE>graphelastic.GraphAnimFrame</CODE>
 * @author <A href=http://www.alapetite.fr.st>Alexandre Alapetite</A>, <A href=http://www.andujar.fr.st>Brice Andujar</A>, <A href=mailto:gregory.gontier@free.fr>Gregory Gontier</A>
 * @version 1.0
 * @see Sommet
 * @see graphelastic.GraphAnimFrame
 */
public class Graphe extends Object implements GrapheInterface
{
  /**
   * Ensemble des sommets de ce graphe.
   * @see Sommet
   */
  Vector vSommets;

  /**
   * Dans le cas d'un affichage graphique, objet qui s'occupe de a.
   */
  private GraphAnimFrame monGrapheAnim;

  /**
   * Indique si un affichage graphique est demand.
   */
  private final boolean graphique;

  /**
   * Lors de la squence de destruction, indique le nombre de threads encore en vie.
   */
  private int nbThreadsRestants;

  private final int nbColorations;

  public Graphe(int nbSommets,int nbColorations,boolean graphique)
  {
    super();
    vSommets = new Vector(nbSommets);
    this.graphique = graphique;
    this.nbColorations = nbColorations;
    if (graphique) monGrapheAnim = new GraphAnimFrame();
    for (int cpt=0; cpt<nbSommets; cpt++)
      addSommet(cpt);
  }

  public Graphe(String fileName,int nbColorations,boolean graphique)
  {
    super();
    vSommets = new Vector();
    this.graphique = graphique;
    this.nbColorations = nbColorations;
    if (graphique)
    {
      monGrapheAnim = new GraphAnimFrame();
      monGrapheAnim.setAuto(false);
    }
    FileXML fileXML = new FileXML();
    fileXML.readFromFile(this,fileName);
  }

  /**
   * Arrte les threads de tous les sommets avec <A href=Sommet.html#destroy()>Sommet.arrete()</A>, et nettoie ce graphe.
   * Les rponses des threads sont attendues  <CODE>desinscription</CODE>
   * et la fin totale de la destruction est  <CODE>finDesinscription</CODE>
   * @see #desinscription
   * @see #finDesinscription
   * @see Sommet#arrete
   */
  public void destroy()
  {
    if (graphique)
    {
      monGrapheAnim.destroy();
      monGrapheAnim = null;
    }
    nbThreadsRestants = vSommets.size();
    for (int cpt=0; cpt<vSommets.size(); cpt++)
      ((Sommet)vSommets.elementAt(cpt)).arrete(this);
    vSommets.clear();
    vSommets = null;
  }

  /**
   * Dmarre les threads de tous les sommets avec <A href=Sommet.html#start()>Sommet.start()</A>
   * @see Sommet#run
   */
  public void start()
  {
    if (graphique)
    {
      monGrapheAnim.start();
      monGrapheAnim.validate();
      monGrapheAnim.setVisible(true);
    }
    for (int cpt=0; cpt<vSommets.size(); cpt++)
      ((Sommet)vSommets.elementAt(cpt)).start();
    //((Sommet)vSommets.elementAt(0)).demarreElection();
    ((Sommet)vSommets.lastElement()).demarreElection();
    ((Sommet)vSommets.get(vSommets.size()/2)).demarreElection();
    ((Sommet)vSommets.firstElement()).demarreElection();
  }

  public void desinscription(Sommet sommet)
  {
    if (--nbThreadsRestants<=0) finDesinscription();
  }

  /**
   * Mthode invoque lorsque tous les threads se sont nettoys et auto-dtruits.
   */
  void finDesinscription()
  {
    affiche("Tous les threads sont morts");
  }

  /**
   * Jeu de test.
   */
  public void test()
  {

  }

  /**
   * Informations sur le nombre de messages changs.
   */
  public void informations()
  {
    affiche("Messages Election : "+SommetElection.nbMessagesElection+
            "\nMessages Arbre Couvrant : "+SommetArbreCouvrant.nbMessagesArbreCouvrant+
            "\nMessages Numerotation Profondeur : "+SommetNumProfondeur.nbMessagesNumProfondeur+
            "\nMessages Coloration : "+SommetColoration.nbMessagesColoration+
            "\nMessages Coloration Asynchone : "+SommetColorationAsynch1.nbMessagesColoration_SJ+
            "\nNombre de couleurs utiles : "+nbCouleursUtiles()+
            "\nBackTracks : "+SommetColoration.nbBackTrack+
            "\nBacTracks Paralleles: "+SommetColorationAsynch1.nbBackTrack_SJ);
  }

  /**
   * Indique si des sommets ont encore des messages  traiter.
   */
  public boolean isActive()
  {
    boolean act = false;
    for (int cpt=vSommets.size()-1; !act && (cpt>=0); cpt--)
      act = !getSommetAt(cpt).isFini();
    return act;
  }

  /**
   * Indique le nombre de couleurs diffrentes utilises dans le graphe.
   */
  final int nbCouleursUtiles()
  {
    BooleanVector coulUt = new BooleanVector(nbColorations,false);
    for (int cpt=vSommets.size()-1; cpt>=0; cpt--)
      coulUt.set(getSommetAt(cpt).getCouleur());
    int nbCoul = 0;
    for (int cpt=coulUt.capacity(); cpt>=0; cpt--)
      if (coulUt.get(cpt)) nbCoul++;
    return nbCoul;
  }

  final void addSommet(int num)
  {
    vSommets.add(new Sommet(nbColorations,num));
  }

  final void addSommet(int num, double x, double y)
  {
    Sommet monSommet = new Sommet(nbColorations,num);
    vSommets.add(monSommet);
    if (graphique) monGrapheAnim.addNode(monSommet,x,y);
  }

  /**
   * Accs  un sommet du graphe.
   */
  final Sommet getSommetAt(int index)
  {
    return ((Sommet)vSommets.elementAt(index));
  }

  /**
   * Gnre des arcs entre les sommets de ce graphe de manire alatoire.
   * @param densite probabilit qu'il y ait une arrte entre deux sommets.
   * @param randomSeed graine pour le gnrateur de nombres alatoires.
   */
  public void genereArretesAlleatoires(double densite, long randomSeed)
  {
    Random monRand = new Random(randomSeed);
    for (int sommet=0; sommet<vSommets.size(); sommet++)
      for (int cpt=0; cpt<sommet; cpt++)
        if (densite>=monRand.nextDouble()) setArrete(sommet,cpt);
  }

  /**
   * Gnre des arcs entre les sommets de ce graphe de manire alatoire.
   * @param densite probabilit qu'il y ait une arrte entre deux sommets.
   */
  public void genereArretesAlleatoires(double densite)
  {
    genereArretesAlleatoires(densite,System.currentTimeMillis());
  }

  /**
   * Relie deux sommets par un arc direct.
   */
  public void setArrete(int sommet1, int sommet2)
  {
    Sommet som1 = ((Sommet)vSommets.elementAt(sommet1));
    Sommet som2 = ((Sommet)vSommets.elementAt(sommet2));
    setArrete(som1,som2);
  }

  /**
   * Relie deux sommets par un arc direct.
   */
  void setArrete(Sommet sommet1, Sommet sommet2)
  {
    if (!sommet1.isVoisin(sommet2.getIdentifiant()))
    {
      sommet1.ajoutVoisin(sommet2.algoDMessageFIFO, sommet2.identifiant);
      sommet2.ajoutVoisin(sommet1.algoDMessageFIFO, sommet1.identifiant);
      if (graphique) monGrapheAnim.addEdge(sommet1,sommet2);
    }
  }

  /**
   * Affiche une ligne de texte sur la sortie standard.
   */
  static void affiche(String texte)
  {
    System.out.println(texte);
  }

  /**
   * Affiche du texte sur la sortie standard.
   */
  static void afficheErreur(String texte)
  {
    System.err.println(texte);
  }

  /**
   * Information sur ce graphe.
   * @return du texte sur plusieurs lignes dcrivant ce graphe et ses sommets.
   */
  public String toString()
  {
    String result="Graphe\n";
    for (int cpt=0; cpt<vSommets.size(); cpt++)
      result+="Sommet "+cpt+" : "+vSommets.elementAt(cpt).toString()+"\n";
    return result;
  }

}
