package algod;

import java.util.Vector;

/**
 * Classe Thread dfinissant les attributs d'un sommet.
 * @author <A href=http://www.alapetite.fr.st>Alexandre Alapetite</A>, <A href=http://www.andujar.fr.st>Brice Andujar</A>, <A href=mailto:gregory.gontier@free.fr>Gregory Gontier</A>
 * @version 1.0
 */
abstract class SommetAbstract extends Thread
{
  /**
   * Indique si un texte doit apparatre lors de certaines actions comme
   * l'envoit de messages.
   */
  public static boolean verbose = true;

  /**
   * Identifiant unique de ce sommet.
   */
  protected final int identifiant;

  /**
   * Tableau des voisins <CODE>SommetVoisin</CODE> immdiats de ce sommet.
   * @see SommetVoisin
   */
  private Vector voisins;

  /**
   * @param identifiant identifiant unique de ce sommet.
   */
  public SommetAbstract(int identifiant)
  {
    super("Sommet "+identifiant);
    this.setPriority((Thread.NORM_PRIORITY+Thread.MIN_PRIORITY)/2); //facultatif
    this.identifiant = identifiant;
    this.voisins = new Vector();
  }

  /**
   * Arrte le thread et nettoie ses structures et moyens de communication. <BR>
   * Attention, les autres threads ne sont pas prvenus de sa destruction.
   */
  public void destroy()
  {
    for (int cpt=nbVoisins()-1; cpt>0; cpt--) getVoisinAt(cpt).destroy();
    voisins.clear();
    voisins = null;
    if (verbose) affiche(""+identifiant+" meurt");
    //super.destroy(); //mthode pas implmente dans Thread
  }

  /**
   * Accs  l'identifiant de ce sommet.
   */
  public final int getIdentifiant()
  {
    return this.identifiant;
  }

  /**
   * Teste si un sommet est voisin de celui-ci.
   * @param ident identifiant du sommet potentiellement voisin de celui-ci.
   * @return si le sommet de cet identifiant est voisin de ce sommet.
   */
  public final boolean isVoisin(int ident)
  {
    for (int cpt=voisins.size()-1; cpt>=0; cpt--)
      if (getVoisinAt(cpt).getIdentifiant() == ident) return true;
    return false;
  }

  /**
   * @param num le numro du voisin auquel on veut avoir accs.
   * @return le num-ime voisin.
   */
  public final SommetVoisin getVoisinAt(int num)
  {
    return ((SommetVoisin)voisins.get(num));
  }

  /**
   * @param ident l'identifiant du voisin auquel on veut avoir accs.
   * @return le voisin de l'identifiant ident, null sinon.
   */
  public final SommetVoisin getVoisin(int ident)
  {
    for (int cpt=voisins.size()-1; cpt>=0; cpt--)
      if (getVoisinAt(cpt).getIdentifiant() == ident) return getVoisinAt(cpt);
    return null;
  }

  /**
   * @param ident l'identifiant d'un sommet voisin de ce sommet.
   * @return le numro du voisin de cet identifiant, -1 s'il n'y a pas de voisin avec cet identifiant.
   */
  public final int numOfVoisin(int ident)
  {
    for (int cpt=voisins.size()-1; cpt>=0; cpt--)
      if (getVoisinAt(cpt).getIdentifiant() == ident) return cpt;
    return -1;
  }

  /**
   * @return le nombre de sommets voisins de celui-ci.
   */
  public final int nbVoisins()
  {
    return voisins.size();
  }

  /**
   * Ajoute un voisin  ce sommet. <BR>
   * Ne met pas  jour les algorithmes dj effectus.
   */
  protected int ajoutVoisin(SommetVoisin unSommet)
  {
    voisins.add(unSommet);
    return nbVoisins()-1;
  }

  /**
   * Ajoute un voisin  ce sommet. <BR>
   * Ne met pas  jour les algorithmes dj effectus.
   */
  protected final int ajoutVoisin(AlgoDMessageFIFO aAlgoDMessageFIFO, int aIdentifiant)
  {
    return ajoutVoisin(new SommetVoisin(aAlgoDMessageFIFO,aIdentifiant));
  }

  /**
   * Affiche du texte sur la sortie standard.
   */
  void affiche(String texte)
  {
    System.out.println(texte);
  }

  /**
   * Affiche du texte sur la sortie standard.
   */
  void afficheErreur(String texte)
  {
    System.err.println(texte);
  }

  /**
   * Fait un beep
   */
  void sonnerie()
  {
    System.out.println((char)(7)); //ASCII(7) c'est un beep en console
  }

  /**
   * Attend un certain nombre de millisecondes.
   */
  final void attend(long millis)
  {
    this.yield();
    try
    {
      this.sleep(millis);
    }
    catch (InterruptedException ie)
    {
      affiche(ie.toString());
    }
    this.yield();
  }

  /**
   * Informations sur ce sommet.
   * @return un texte sur une ligne dcrivant de sommet.
   */
  public String toString()
  {
    String result="ident="+identifiant+"; voisins:";
    int nbVois = nbVoisins();
    for (int cpt=0; cpt<nbVois; cpt++)
     result+=" "+getVoisinAt(cpt).getIdentifiant();
    return result;
  }

}
