package algod;

import java.util.Vector;
//import java.math.BigInteger;

/**
 * Implmente un arbre couvrant de la composante connexe d'un sommet.
 * @author <A href=http://www.alapetite.fr.st>Alexandre Alapetite</A>, <A href=http://www.andujar.fr.st>Brice Andujar</A>, <A href=mailto:gregory.gontier@free.fr>Gregory Gontier</A>
 * @version 1.0
 */
class SommetArbreCouvrant extends SommetElection
{
  static int nbMessagesArbreCouvrant = 0;

  /**
   * Valeur indiquant que le sommet est racine de l'arbre couvrant.
   * @see #numVoisinParent
   */
  private static final int CVR_RACINE = -1;

  /**
   * Valeur indiquant que le sommet ne connait pas encore son pre.
   * @see #numVoisinParent
   */
  private static final int CVR_PARENT_UNKNOWN = -2;

  /**
   * Tableau des enfants <CODE>SommetInterface</CODE> de ce sommet pour un arbre couvrant.
   * @see #voisins
   * @see SommetVoisin
   */
  private Vector enfants;

  /**
   * Numro du voisin qui fait parent.
   */
  int numVoisinParent;

  /**
   * Poids de ce noeud de l'arbre recouvrant,
   * c'est--dire le nombre de sommets enfants et petit-enfants, etc...
   */
  protected int poids;

  /**
   * Tableau stoquant les poids des diffrents enfants.
   * Utilis pour renumroter les sommets.
   */
  IntVector poidsEnfants;

  /**
   * Nombre de rponses d'enfants potentiels restantes.
   */
  private int nbReponsesEnfantsRestantes;

  /**
   * Nombre de rponses restantes indiquant que l'algo d'arbre couvrant est termin.
   */
  private int nbReponsesFiniRestantes;

  public SommetArbreCouvrant(int identifiant)
  {
    super(identifiant);
    this.enfants = new Vector();
    this.numVoisinParent = CVR_PARENT_UNKNOWN;
    this.poids = 0;
    this.poidsEnfants = new IntVector();
    this.nbReponsesEnfantsRestantes = Integer.MAX_VALUE;
    this.nbReponsesFiniRestantes = Integer.MAX_VALUE;
    //this.identifiantFiliation = BigInteger.ZERO;
  }

  public void destroy()
  {
    //identifiantFiliation = null;
    enfants.clear();
    enfants = null;
    poidsEnfants.destroy();
    poidsEnfants = null;
    super.destroy();
  }

  /**
   * Indique si ce sommet est la racine de l'arbre couvrant.
   */
  public final boolean isRacine()
  {
    return numVoisinParent == CVR_RACINE;
  }

  /**
   * Indique si ce sommet est une feuille de l'arbre couvrant.
   */
  public final boolean isFeuille()
  {
    return enfants.size() == 0;
  }

  /**
   * Indique si ce sommet est un fils du sommet de cet identifiant.
   */
  public final boolean isFilsDe(int idPere)
  {
    SommetVoisin sv = this.getPere();
    return (sv!=null) && (sv.getIdentifiant() == idPere);
  }

  /**
   * @return le nombre d'enfants de ce sommet dans l'arbre couvrant.
   */
  public final int nbEnfants()
  {
    return enfants.size();
  }

  /**
   * @param num le numro de l'enfant auquel on veut avoir accs.
   * @return le num-ime enfant.
   */
  public final SommetVoisin getEnfantAt(int num)
  {
    return (SommetVoisin)enfants.elementAt(num);
  }

  /**
   * Accs au pre de ce sommet.
   */
  public final SommetVoisin getPere()
  {
    return (numVoisinParent<0) ? null : getVoisinAt(numVoisinParent);
  }

  /**
   * @param ident l'identifiant du sommet enfant auquel on veut avoir accs.
   * @return l'enfant de l'identifiant ident, null sinon.
   */
  public final SommetVoisin getEnfant(int ident)
  {
    for (int cpt=enfants.size()-1; cpt>=0; cpt--)
      if (getEnfantAt(cpt).getIdentifiant() == ident) return getEnfantAt(cpt);
    return null;
  }

  /**
   * @param ident l'identifiant du sommet enfant auquel on veut avoir accs.
   * @return le numro d'enfant de ce sommet voisin, -1 s'il n'est pas enfant.
   */
  public int indexOfEnfant(int ident)
  {
    for (int cpt=enfants.size()-1; cpt>=0; cpt--)
      if (getEnfantAt(cpt).getIdentifiant() == ident) return cpt;
    return -1;
  }

  /**
   * Ajoute un enfant  ce sommet.
   */
  private void ajoutEnfant(SommetVoisin unSommet)
  {
    enfants.add(unSommet);
  }

  /**
   * A appeler pour dmarrer l'algorithme distribu d'arbre couvrant  partir de ce sommet. <BR>
   * Ce sommet sera racine.
   */
  public void demarreArbreCouvrant()
  {
    affiche("--- "+identifiant+" debutte l'arbre couvrant-------------------");
    numVoisinParent = CVR_RACINE;
    nbReponsesEnfantsRestantes = nbVoisins();
    nbReponsesFiniRestantes = nbReponsesEnfantsRestantes;
    long premierDispo = 1;
    int maxim = nbVoisins();
    envoyerATousVoisins(AlgoDMessage.MSG_ARBCOUV_PARENT,"",identifiant); //on cherche ses enfants
  }

  /**
   * Mthode invoque lorsque l'algorithme d'arbre couvrant est termin.<BR>
   * Ce sommet est racine.
   */
  protected void finArbreCouvrant()
  {
    affiche("--- "+identifiant+" a fini l'arbre couvrant (poids "+(poids+1)+") ---");
    attend(10);
  }

  /**
   * Termine la mthode hrite de fin d'lection pour dmarrer l'arbre couvrant.
   */
  protected final void finElection()
  {
    super.finElection();
    demarreArbreCouvrant();
  }

  /**
   * Envoie un message au pre de ce sommet dans l'arbre couvrant.
   */
  public void envoyerAPere(char typeMessage, String contenu, int identSommetSource)
  {
    envoyerAVoisin(typeMessage,contenu,identSommetSource,numVoisinParent);
  }

  /**
   * Envoie un message aux enfants de ce sommet dans l'arbre couvrant.
   */
  public void envoyerAuxEnfants(char typeMessage, String contenu, int identSommetSource)
  {
    for (int cpt=nbEnfants()-1; cpt>=0; cpt--)
      envoyerASommet(typeMessage,contenu,identSommetSource,getEnfantAt(cpt));
  }

  /**
   * Envoie un message  un enfant.
   * @param numEnfant le n de l'enfant voisin  qui ce message est destin.
   */
  public void envoyerAEnfant(char typeMessage, String contenu, int identSommetSource, int numEnfant)
  {
    envoyerASommet(typeMessage,contenu,identSommetSource,getEnfantAt(numEnfant));
  }

  protected void traiteReception(AlgoDMessage adm, int numVoisin)
  {
    switch (adm.typeMessage)
    {
      case AlgoDMessage.MSG_ARBCOUV_PARENT :
        nbMessagesArbreCouvrant++;
        receptionParent(numVoisin);
        break;
      case AlgoDMessage.MSG_ARBCOUV_ENFANT_OUI :
        nbMessagesArbreCouvrant++;
        receptionEnfantOUI(numVoisin);
        break;
      case AlgoDMessage.MSG_ARBCOUV_ENFANT_NON :
        nbMessagesArbreCouvrant++;
        receptionEnfantNON();
        break;
      case AlgoDMessage.MSG_ARBCOUV_FINI :
        nbMessagesArbreCouvrant++;
        receptionArbreFini(indexOfEnfant(adm.identVoisin),Integer.parseInt(adm.contenu));
        break;
      default : super.traiteReception(adm, numVoisin);
    }
  }

  /**
   * Reception d'une candidature de pre.
   */
  private void receptionParent(int numVoisin)
  {
    if (numVoisinParent == CVR_PARENT_UNKNOWN) //si l'opration d'arbre couvrant n'a pas dj t faite
    {
      numVoisinParent = numVoisin;
      envoyerAPere(AlgoDMessage.MSG_ARBCOUV_ENFANT_OUI,"",identifiant); //on s'enregistre comme enfant auprs de notre pre
      nbReponsesEnfantsRestantes = nbVoisins()-1;
      nbReponsesFiniRestantes = nbReponsesEnfantsRestantes;
      if (nbReponsesEnfantsRestantes<=0) receptionEnfantNON(); //on est forcemment une feuille car on a un seul voisin (ou aucun!).
      else envoyerATousVoisinsSauf(AlgoDMessage.MSG_ARBCOUV_PARENT,"",identifiant,numVoisinParent); //on cherche ses enfants
    }
    else envoyerAVoisin(AlgoDMessage.MSG_ARBCOUV_ENFANT_NON,"",identifiant,numVoisin); //on a dj un pre, on refuse cette candidature
  }

  /**
   * Reception d'un enfant d'accord.
   */
  private void receptionEnfantOUI(int numVoisin)
  {
    ajoutEnfant(getVoisinAt(numVoisin));
    nbReponsesEnfantsRestantes--;
  }

  /**
   * Reception d'un enfant pas d'accord.
   */
  private void receptionEnfantNON()
  {
    nbReponsesEnfantsRestantes--;
    nbReponsesFiniRestantes--; //on aura pas de fin d'algo  attendre de la part de ce voisin
    //si on est une feuille,
    //ou que le dernier voisin dont on attendait la rponse dit non pour tre enfant
    //et que tous les autres enfants ont dj dit qu'ils avaient fini,
    //on lance le code de fin artificiellement
    if (nbReponsesFiniRestantes<=0)
      receptionArbreFini(-1,0);
  }

  /**
   * Rception par un enfant que les sous-arbre de cet enfant est termin.
   */
  private void receptionArbreFini(int numEnfant, int SousPoids)
  {
    nbReponsesFiniRestantes--;
    if (numEnfant>=0) poidsEnfants.set(numEnfant,SousPoids); //stocke les poids des diffrents enfants
    this.poids+=SousPoids;
    if (nbReponsesFiniRestantes<=0)
    {
      if (isRacine()) finArbreCouvrant();
      else envoyerAPere(AlgoDMessage.MSG_ARBCOUV_FINI,""+(poids+1),identifiant);
    }
  }

  public String toString()
  {
    String result =  super.toString()+"; poids:"+poids+"; enfants:";
    int nbEnf = nbEnfants();
    for (int cpt=0; cpt<nbEnf; cpt++)
      result+=" "+getEnfantAt(cpt).getIdentifiant();
    return result;
  }

}
